mod modified_parsers;
mod primitive_parsers;

use crate::parser::modified_parsers::{
    BindParser, ConvertParser, LitervalParser, LookAheadParser, MapParser, NextParser,
    ReverseParser,
};
use crate::parser::primitive_parsers::{
    AnyParser, FailParser, FailWithMessageParser, SatisfyParser, SkipParser, SucceedParser,
    TakeParser,
};
use std::cell::RefCell;
use std::fmt::{Debug, Display, Formatter};
use std::marker::PhantomData;
use std::rc::Rc;

#[derive(Debug)]
pub struct FailedParserResult {
    message: String,
}

impl FailedParserResult {
    pub fn new(message: String) -> Self {
        Self { message }
    }
}

impl Display for FailedParserResult {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        write!(f, "Parse failed: {}.", self.message)
    }
}

pub type ParserResult<'a, TToken, T> =
    Result<(&'a [TToken], T), (&'a [TToken], FailedParserResult)>;

pub struct ParserContext<TToken, TContext>
where
    TToken: Debug + Clone,
{
    pub context: TContext,
    input_array: Vec<TToken>,
}

impl<TToken, TContext> ParserContext<TToken, TContext>
where
    TToken: Debug + Clone,
{
    pub fn new(input: &[TToken], context: TContext) -> Rc<RefCell<Self>> {
        Rc::new(RefCell::new(Self {
            input_array: input.iter().map(|x| x.clone()).collect(),
            context,
        }))
    }

    pub fn input_slice(&self) -> &[TToken] {
        self.input_array.as_slice()
    }
}

impl<TContext> ParserContext<char, TContext> {
    pub fn new_with_str(input: &str, context: TContext) -> Rc<RefCell<Self>> {
        Rc::new(RefCell::new(Self {
            input_array: input.chars().collect(),
            context,
        }))
    }
}

pub trait Parser<TToken, T, TContext>
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T>;

    fn map<TResult, F>(self, f: F) -> MapParser<Self, F, T>
    where
        Self: Sized,
        F: Fn(&T) -> TResult,
    {
        MapParser {
            parser: self,
            mapper: f,
            phantom: PhantomData,
        }
    }

    fn bind<TResult, F, P>(self, f: F) -> BindParser<Self, F, T>
    where
        Self: Sized,
        F: Fn(&T) -> P,
        P: Parser<TToken, TResult, TContext>,
    {
        BindParser {
            parser: self,
            binder: f,
            phantom_data: PhantomData,
        }
    }

    fn convert<TResult>(self) -> ConvertParser<Self, T, TResult>
    where
        Self: Sized,
    {
        ConvertParser {
            parser: self,
            phantom_data1: PhantomData,
            phantom_data2: PhantomData,
        }
    }

    fn next<TResult, H>(self, handler: H) -> NextParser<Self, H, T>
    where
        Self: Sized,
        H: for<'f> Fn(ParserResult<'f, TToken, T>) -> ParserResult<'f, TToken, TResult>,
    {
        NextParser {
            parser: self,
            handler,
            phantom_data: PhantomData,
        }
    }

    fn literal(self) -> LitervalParser<Self, T>
    where
        Self: Sized,
    {
        LitervalParser {
            parser: self,
            phantom_data: PhantomData,
        }
    }

    fn look_ahead(self) -> LookAheadParser<Self, T>
    where
        Self: Sized,
    {
        LookAheadParser {
            parser: self,
            phantom_data: PhantomData,
        }
    }

    fn reverse<TResult>(self, result: TResult) -> ReverseParser<Self, T, TResult>
    where
        Self: Sized,
        TResult: Debug + Clone,
    {
        ReverseParser {
            parser: self,
            result,
            phantom_data: PhantomData,
        }
    }
}

impl<TToken, T, TContext, F> Parser<TToken, T, TContext> for F
where
    TToken: Debug + Clone,
    F: for<'f> Fn(
        Rc<RefCell<ParserContext<TToken, TContext>>>,
        &'f [TToken],
    ) -> ParserResult<'f, TToken, T>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        (*self)(context, input)
    }
}

pub fn succeed<TToken, T, TContext>(value: T) -> impl Parser<TToken, T, TContext>
where
    TToken: Debug + Clone,
    T: Debug + Clone,
{
    SucceedParser { value }
}

pub fn fail<TToken, T, TContext>() -> impl Parser<TToken, T, TContext>
where
    TToken: Debug + Clone,
{
    FailParser {
        phantom_data: PhantomData,
    }
}

pub fn fail_with_message<TToken, T, TContext>(message: &str) -> impl Parser<TToken, T, TContext>
where
    TToken: Debug + Clone,
{
    FailWithMessageParser {
        message: message.to_owned(),
        phantom_data: PhantomData,
    }
}

pub fn satisfy<TToken, TContext, TP>(predicate: TP) -> impl Parser<TToken, TToken, TContext>
where
    TToken: Debug + Clone,
    TP: Fn(&TToken) -> bool,
{
    SatisfyParser { predicate }
}

pub fn any<TToken, TContext>() -> impl Parser<TToken, TToken, TContext>
where
    TToken: Debug + Clone,
{
    AnyParser {}
}

pub fn take<TToken, TContext>(count: usize) -> impl Parser<TToken, Vec<TToken>, TContext>
where
    TToken: Debug + Clone,
{
    TakeParser { count }
}

pub fn skip<TToken, TContext>(count: usize) -> impl Parser<TToken, (), TContext>
where
    TToken: Debug + Clone,
{
    SkipParser { count }
}

#[cfg(test)]
pub fn parser_test_helper<T, P>(
    context: Rc<RefCell<ParserContext<char, ()>>>,
    excepted_node: &T,
    test_parser: P,
) where
    T: PartialEq + Debug,
    P: Parser<char, T, ()>,
{
    let borrowed_context = context.borrow();
    let input = borrowed_context.input_slice();
    let (_, actual_result) = test_parser.parse(context.clone(), input).unwrap();

    assert_eq!(excepted_node, &actual_result);
}

#[cfg(test)]
pub fn failed_parser_test_helper<T, P>(
    context: Rc<RefCell<ParserContext<char, ()>>>,
    test_parser: P,
) where
    T: Debug,
    P: Parser<char, T, ()>,
{
    let borrowed_context = context.borrow();
    let input = borrowed_context.input_slice();
    let result = test_parser.parse(context.clone(), input);

    assert!(result.is_err());
}
