use crate::parser::{FailedParserResult, Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct MapParser<P, F, T1> {
    pub(crate) parser: P,
    pub(crate) mapper: F,
    pub(crate) phantom: PhantomData<T1>,
}

impl<TToken, T1, T2, TContext, P, F> Parser<TToken, T2, TContext> for MapParser<P, F, T1>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T1, TContext>,
    F: Fn(&T1) -> T2,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        self.parser
            .parse(context, input)
            .map(|(remainder, result)| (remainder, (self.mapper)(&result)))
    }
}

pub struct BindParser<P, F, T1> {
    pub(crate) parser: P,
    pub(crate) binder: F,
    pub(crate) phantom_data: PhantomData<T1>,
}

impl<TToken, T1, T2, TContext, P, P2, F> Parser<TToken, T2, TContext> for BindParser<P, F, T1>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T1, TContext>,
    F: Fn(&T1) -> P2,
    P2: Parser<TToken, T2, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        let (input, middle) = self.parser.parse(context.clone(), input)?;

        (self.binder)(&middle).parse(context, input)
    }
}

pub struct ConvertParser<P, T1, T2> {
    pub(crate) parser: P,
    pub(crate) phantom_data1: PhantomData<T1>,
    pub(crate) phantom_data2: PhantomData<T2>,
}

impl<TToken, T1, T2, TContext, P> Parser<TToken, T2, TContext> for ConvertParser<P, T1, T2>
where
    TToken: Debug + Clone,
    T2: From<T1>,
    P: Parser<TToken, T1, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T2> {
        self.parser
            .parse(context, input)
            .map(|(input, result)| (input, result.into()))
    }
}

pub struct NextParser<P, H, T> {
    pub(crate) parser: P,
    pub(crate) handler: H,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TResult, TContext, P, H> Parser<TToken, TResult, TContext> for NextParser<P, H, T>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
    H: for<'f> Fn(ParserResult<'f, TToken, T>) -> ParserResult<'f, TToken, TResult>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TResult> {
        let result = self.parser.parse(context, input);

        (self.handler)(result)
    }
}

pub struct LitervalParser<P, T> {
    pub(crate) parser: P,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TContext, P> Parser<TToken, Vec<TToken>, TContext> for LitervalParser<P, T>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<TToken>> {
        let origin_input = input;
        let (input, _) = self.parser.parse(context, input)?;

        if origin_input.is_empty() {
            return Err((
                input,
                FailedParserResult::new(
                    "Empty input, the 'literal' parser cann't get length of elememtn.".to_owned(),
                ),
            ));
        }

        let length = (&origin_input[1..]).as_ptr() as usize - origin_input.as_ptr() as usize;
        let index = (input.as_ptr() as usize - origin_input.as_ptr() as usize) / length;

        Ok((
            input,
            (&origin_input[..index])
                .into_iter()
                .map(|x| x.clone())
                .collect(),
        ))
    }
}

pub struct LookAheadParser<P, T> {
    pub(crate) parser: P,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TContext, P> Parser<TToken, T, TContext> for LookAheadParser<P, T>
where
    TToken: Debug + Clone,
    P: Parser<TToken, T, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        let original_input = input;
        let (_, r) = self.parser.parse(context, input)?;

        Ok((original_input, r))
    }
}

pub struct ReverseParser<P, T, TResult> {
    pub(crate) parser: P,
    pub(crate) result: TResult,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TResult, TContext, P> Parser<TToken, TResult, TContext>
    for ReverseParser<P, T, TResult>
where
    TToken: Debug + Clone,
    TResult: Debug + Clone,
    P: Parser<TToken, T, TContext>,
{
    fn parse<'a>(
        &self,
        context: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TResult> {
        match self.parser.parse(context, input) {
            Ok((input, _)) => Err((
                input,
                FailedParserResult::new("Reverse failed succeeded.".to_owned()),
            )),
            Err((input, _)) => Ok((input, self.result.clone())),
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        failed_parser_test_helper, parser_test_helper, satisfy, take, FailedParserResult,
    };
    use crate::text::char_parser;

    #[test]
    fn map_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            take(5).map(|_| ()),
        )
    }

    #[test]
    fn bind_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'b',
            char_parser('a').bind(|_| char_parser('b')),
        );

        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'c',
            char_parser('a')
                .bind(|_| char_parser('b'))
                .bind(|_| char_parser('c')),
        );
    }

    struct Number(i32);

    impl From<Number> for String {
        fn from(value: Number) -> Self {
            format!("{}", value.0)
        }
    }

    #[test]
    fn convert_test() {
        fn single_numer_parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, Number> {
            let (input, result) = satisfy(|c: &char| c.is_numeric()).parse(context, input)?;

            match result.to_digit(10) {
                None => Err((input, FailedParserResult::new("What?".to_string()))),
                Some(r) => Ok((input, Number(r as i32))),
            }
        }

        parser_test_helper(
            ParserContext::new_with_str("9", ()),
            &"9".to_owned(),
            single_numer_parser.convert(),
        );

        parser_test_helper(
            ParserContext::new_with_str("1", ()),
            &"1".to_owned(),
            single_numer_parser.convert(),
        );

        failed_parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            single_numer_parser.convert::<String>(),
        );
    }

    #[test]
    fn next_test() {
        let parser = || {
            satisfy(|c: &char| c.is_numeric()).next(|r| match r {
                Ok((input, result)) => match result.to_digit(10) {
                    None => Err((input, FailedParserResult::new("What?".to_string()))),
                    Some(r) => Ok((input, Number(r as i32))),
                },
                Err(r) => Err(r),
            })
        };

        parser_test_helper(
            ParserContext::new_with_str("9", ()),
            &"9".to_owned(),
            parser().convert(),
        );

        parser_test_helper(
            ParserContext::new_with_str("1", ()),
            &"1".to_owned(),
            parser().convert(),
        );

        failed_parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            parser().convert::<String>(),
        );
    }

    #[test]
    fn literal_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec!['a'],
            char_parser('a').literal(),
        );

        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec!['a', 'b'],
            char_parser('a').bind(|_| char_parser('b')).literal(),
        );

        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &vec!['a', 'b', 'c'],
            char_parser('a')
                .bind(|_| char_parser('b'))
                .bind(|_| char_parser('c'))
                .literal(),
        );
    }

    #[test]
    fn look_ahead_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            char_parser('a').look_ahead().bind(|_| char_parser('a')),
        );
    }

    #[test]
    fn reverse_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &(),
            char_parser('b').reverse(()),
        );

        failed_parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            char_parser('a').reverse(()),
        );
    }
}
