use crate::parser::{FailedParserResult, Parser, ParserContext, ParserResult};
use std::cell::RefCell;
use std::fmt::Debug;
use std::marker::PhantomData;
use std::rc::Rc;

pub struct SucceedParser<T> {
    pub(crate) value: T,
}

impl<TToken, T, TContext> Parser<TToken, T, TContext> for SucceedParser<T>
where
    TToken: Debug + Clone,
    T: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Ok((input, self.value.clone()))
    }
}

pub struct FailParser<T> {
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TContext> Parser<TToken, T, TContext> for FailParser<T>
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Err((
            input,
            FailedParserResult::new("Default failed parser.".to_owned()),
        ))
    }
}

pub struct FailWithMessageParser<T> {
    pub(crate) message: String,
    pub(crate) phantom_data: PhantomData<T>,
}

impl<TToken, T, TContext> Parser<TToken, T, TContext> for FailWithMessageParser<T>
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, T> {
        Err((input, FailedParserResult::new(self.message.clone())))
    }
}

pub struct SatisfyParser<F> {
    pub(crate) predicate: F,
}

impl<TToken, TContext, F> Parser<TToken, TToken, TContext> for SatisfyParser<F>
where
    TToken: Debug + Clone,
    F: Fn(&TToken) -> bool,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TToken> {
        if input.is_empty() {
            return Err((input, FailedParserResult::new("Input is empty.".to_owned())));
        }

        if (self.predicate)(&input[0]) {
            Ok((&input[1..], input[0].clone()))
        } else {
            Err((
                input,
                FailedParserResult::new("Predicate failed.".to_owned()),
            ))
        }
    }
}

pub struct AnyParser {}

impl<TToken, TContext> Parser<TToken, TToken, TContext> for AnyParser
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, TToken> {
        if input.is_empty() {
            Err((input, FailedParserResult::new("Input is empty.".to_owned())))
        } else {
            Ok((&input[1..], input[0].clone()))
        }
    }
}

pub struct TakeParser {
    pub(crate) count: usize,
}

impl<TToken, TContext> Parser<TToken, Vec<TToken>, TContext> for TakeParser
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, Vec<TToken>> {
        if input.len() < self.count {
            Err((
                input,
                FailedParserResult::new(format!(
                    "The input doesn't contain enough {} elemements.",
                    self.count
                )),
            ))
        } else {
            Ok((
                &input[self.count..],
                (&input[..self.count]).iter().map(|x| x.clone()).collect(),
            ))
        }
    }
}

pub struct SkipParser {
    pub(crate) count: usize,
}

impl<TToken, TContext> Parser<TToken, (), TContext> for SkipParser
where
    TToken: Debug + Clone,
{
    fn parse<'a>(
        &self,
        _: Rc<RefCell<ParserContext<TToken, TContext>>>,
        input: &'a [TToken],
    ) -> ParserResult<'a, TToken, ()> {
        if input.len() < self.count {
            Err((
                input,
                FailedParserResult::new(format!(
                    "The input doesn't contain enough {} elemements.",
                    self.count
                )),
            ))
        } else {
            Ok((&input[self.count..], ()))
        }
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        any, fail, fail_with_message, failed_parser_test_helper, parser_test_helper, satisfy, skip,
        succeed, take,
    };

    #[test]
    fn succeed_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &(), succeed(()));
    }

    #[test]
    fn fail_test() {
        failed_parser_test_helper::<(), _>(ParserContext::new_with_str("abc", ()), fail());
    }

    #[test]
    fn fail_with_message_test() {
        failed_parser_test_helper::<(), _>(
            ParserContext::new_with_str("abc", ()),
            fail_with_message("Failed!"),
        );
    }

    #[test]
    fn satisfy_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            satisfy(|x| x == &'a'),
        );

        failed_parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            satisfy(|x| x == &'b'),
        );
    }

    #[test]
    fn any_test() {
        parser_test_helper(ParserContext::new_with_str("abc", ()), &'a', any());

        parser_test_helper(ParserContext::new_with_str("cde", ()), &'c', any());
    }

    #[test]
    fn take_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &("hello".chars().collect()),
            take(5),
        );

        failed_parser_test_helper(ParserContext::new_with_str("abcd", ()), take(5));
    }

    #[test]
    fn skip_test() {
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            skip(5),
        );

        failed_parser_test_helper(ParserContext::new_with_str("abcd", ()), skip(5));
    }
}
