use crate::parser::{satisfy, take, FailedParserResult, Parser};

pub fn char_parser<TContext>(c: char) -> impl Parser<char, char, TContext> {
    satisfy(move |x| *x == c)
}

pub fn string_parser<TContext>(str: String) -> impl Parser<char, String, TContext> {
    take::<char, TContext>(str.len()).next(move |result| {
        result.and_then(|(input, chars)| {
            let chars: String = chars.into_iter().collect();

            if chars == str {
                Ok((input, chars))
            } else {
                Err((
                    input,
                    FailedParserResult::new(format!("Failed to parse '{}'.", str)),
                ))
            }
        })
    })
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::parser::{
        failed_parser_test_helper, parser_test_helper, ParserContext, ParserResult,
    };
    use std::cell::RefCell;
    use std::rc::Rc;

    #[test]
    fn char_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("abc", ()),
            &'a',
            char_parser('a'),
        );

        failed_parser_test_helper(ParserContext::new_with_str("bc", ()), char_parser('a'));
    }

    #[test]
    fn string_parser_test() {
        parser_test_helper(
            ParserContext::new_with_str("Hello, world!", ()),
            &"Hello".to_owned(),
            string_parser("Hello".to_owned()),
        );

        fn test_parser(
            context: Rc<RefCell<ParserContext<char, ()>>>,
            input: &[char],
        ) -> ParserResult<char, String> {
            let (input, first) =
                string_parser("hello, ".to_string()).parse(context.clone(), input)?;
            let (input, second) =
                string_parser("world!".to_owned()).parse(context.clone(), input)?;

            Ok((input, first + second.as_str()))
        }

        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &"hello, world!".to_owned(),
            test_parser,
        );
        parser_test_helper(
            ParserContext::new_with_str("hello, world!", ()),
            &(),
            test_parser.map(|_| ()),
        )
    }
}
